from enum import StrEnum
from typing import Dict


class NEN99971_version(StrEnum):
    V2017 = "2017"
    V2025 = "2025"


class CUR236_version(StrEnum):
    V2023 = "2023"


class Nen99971:
    """
    NEN 9997-1 standard for pile design.
    """

    def __init__(self, version: NEN99971_version = NEN99971_version.V2025) -> None:
        if not isinstance(version, NEN99971_version):
            raise TypeError("version must be a NEN99971_version enum member")
        self._version = version

    def __str__(self) -> str:
        return f"NEN9997-1 ({self._version})"

    @property
    def version(self) -> NEN99971_version:
        """Returns the version of the NEN 9997-1 standard."""
        return self._version


class Cur236:
    """
    CUR236 standard for pile design.
    """

    def __init__(self, version: CUR236_version = CUR236_version.V2023) -> None:
        if not isinstance(version, CUR236_version):
            raise TypeError("version must be a CUR236_version enum member")
        self._version = version

    def __str__(self) -> str:
        return f"CUR236 ({self._version})"

    @property
    def version(self) -> CUR236_version:
        """Returns the version of the CUR236 standard."""
        return self._version


class Norms:
    """
    Container for all norm standards.
    """

    def __init__(
        self,
        nen_9997_1: str = "2025",
        cur_236: str = "2023",
    ) -> None:
        try:
            self.nen_9997_1 = Nen99971(version=NEN99971_version(nen_9997_1))
        except ValueError:
            raise ValueError(
                f"Invalid NEN9997-1 version selected: {nen_9997_1}, must be one of {[v.value for v in NEN99971_version]}"
            )
        try:
            self.cur_236 = Cur236(version=CUR236_version(cur_236))
        except ValueError:
            raise ValueError(
                f"Invalid CUR-236 version selected: {cur_236}, must be one of {[v.value for v in CUR236_version]}"
            )

    def serialize_payload(self) -> Dict[str, str]:
        """
        Serialize the norms to a dictonary payload for the API.

        Returns
        -------
        Dict[str, str]
            A dictionary payload containing the versions of the selected norms.
        """
        payload: Dict[str, str] = {
            "NEN99971_version": self.nen_9997_1.version,
            "CUR236_version": self.cur_236.version,
        }

        return payload



from __future__ import annotations  # noqa: F404

from typing import Any

import numpy as np
import pandas as pd
from IPython.display import DisplayHandle, clear_output, display
from ipywidgets import widgets
from matplotlib import pyplot as plt
from natsort import natsorted

from pypilecore.results.typing import CasesMultiCPTResultsLike


class ViewerCptResultsOverview:
    """
    Viewer for the CPT bearing results overview.

    It offers the following layout:
        - Dropdown widgets:
            - Case: to select the case to show.
            - CPT: to select the CPT to show.
        - Figure Bearing Overview (non-interactive):
    """

    def __init__(self, results_cases: CasesMultiCPTResultsLike) -> None:
        """Initialize the viewer.

        Parameters
        ----------
        results_cases : CasesMultiCPTResultsLike
            The results of the bearing capacity calculations.

        Raises
        ------
        TypeError
            If 'cases_multi_results' are not of type 'CasesMultiCPTResultsLike'.
        """

        # Initialize figure CPT resuls vs. pile tip level
        self.results_cases = results_cases

        # Set up control widgets
        self._case_dropdown = widgets.Dropdown(
            description="Case:",
            value=self.results_cases.cases[0],
            options=self.results_cases.cases,
        )

        _options = natsorted(pd.unique(np.array(self.results_cases.test_ids)))
        self._cpt_dropdown = widgets.Dropdown(
            description="CPT:",
            value=_options[0],
            options=_options,
        )

        # Initiate matplotlib figure as Output widget
        self.plot_widget = widgets.Output()
        self._update_case_and_result(None)  # Initial plot

        # Set up callbacks
        self._case_dropdown.observe(self._update_case_and_result, "value")
        self._cpt_dropdown.observe(self._update_case_and_result, "value")

        # Set up layout
        self._control_widgets = widgets.HBox(
            [
                self._case_dropdown,
                self._cpt_dropdown,
            ]
        )
        self._layout = widgets.VBox(
            [self._control_widgets, self.plot_widget]
        )  # , width=800)

    def _update_case_and_result(self, change: Any) -> None:
        """Private method to update the figure when the case or result name are changed in the control widgets."""

        with self.plot_widget:
            clear_output(wait=True)
            plt.ioff()  # Turn interactive plotting off
            fig = (
                self.results_cases.results_per_case[self._case_dropdown.value]
                .cpt_results[self._cpt_dropdown.value]
                .plot_bearing_overview()
            )
            plt.ion()  # Turn interactive plotting back on
            display(fig)
            plt.close(fig)  # Close the figure to prevent accumulation

    def display(self) -> DisplayHandle | None:
        """Display the figure."""
        return display(self._layout)



from enum import StrEnum
from typing import List


class FrictionRangeStrategy(StrEnum):
    lower_bound = "lower_bound"
    settlement_driven = "settlement_driven"
    manual = "manual"


class SchemaKeys(StrEnum):
    friction_settings = "friction_settings"
    friction_range_strategy = "friction_range_strategy"
    negative_friction = "negative_friction"
    negative_friction_range_nap = "negative_friction_range_nap"
    positive_friction_range_nap = "positive_friction_range_nap"


class FrictionSettings:
    def __init__(
        self,
        friction_range_strategy: (
            str | FrictionRangeStrategy
        ) = FrictionRangeStrategy.lower_bound,
        negative_friction: float | None = None,
        negative_friction_range_nap: List[float] | None = None,
        positive_friction_range_nap: List[float] | None = None,
    ):
        # With loop for compatibility in python 3.11
        assert any(
            item.value == friction_range_strategy for item in FrictionRangeStrategy
        ), f"Invalid value for friction_range_strategy defined, please enter one of {[item.value for item in FrictionRangeStrategy]}"
        self.friction_range_strategy = friction_range_strategy
        self.negative_friction = negative_friction
        self.negative_friction_range_nap = negative_friction_range_nap
        self.positive_friction_range_nap = positive_friction_range_nap

    def serialize_payload(self) -> dict:
        return dict((key, val) for key, val in vars(self).items() if val is not None)



from __future__ import annotations

from enum import StrEnum
from typing import Dict, Hashable, List

from natsort import natsorted
from pygef.common import Location

from pypilecore.results.data_tables import CptResultsTable
from pypilecore.results.grouper_result import GrouperResults
from pypilecore.results.lib import Unit
from pypilecore.results.result_definitions import GrouperResultsDefinition


class ClusterResultColumn(StrEnum):
    characteristic_bearing_capacity = "characteristic_bearing_capacity"
    design_bearing_capacity = "design_bearing_capacity"
    design_negative_friction = "design_negative_friction"
    group_centre_to_centre_validation = "group_centre_to_centre_validation"
    group_centre_to_centre_validation_15 = "group_centre_to_centre_validation_15"
    group_centre_to_centre_validation_20 = "group_centre_to_centre_validation_20"
    group_centre_to_centre_validation_25 = "group_centre_to_centre_validation_25"
    mean_calculated_bearing_capacity = "mean_calculated_bearing_capacity"
    min_calculated_bearing_capacity = "min_calculated_bearing_capacity"
    net_design_bearing_capacity = "net_design_bearing_capacity"
    nominal_cpt = "nominal_cpt"
    variation_coefficient = "variation_coefficient"
    xi_factor = "xi_factor"
    xi_values = "xi_values"


_CLUSTER_RESULT_UNITS = {
    ClusterResultColumn.characteristic_bearing_capacity.value: Unit.KN,
    ClusterResultColumn.design_bearing_capacity.value: Unit.KN,
    ClusterResultColumn.design_negative_friction.value: Unit.KN,
    ClusterResultColumn.group_centre_to_centre_validation.value: Unit.NONE,
    ClusterResultColumn.group_centre_to_centre_validation_15.value: Unit.NONE,
    ClusterResultColumn.group_centre_to_centre_validation_20.value: Unit.NONE,
    ClusterResultColumn.group_centre_to_centre_validation_25.value: Unit.NONE,
    ClusterResultColumn.mean_calculated_bearing_capacity.value: Unit.KN,
    ClusterResultColumn.min_calculated_bearing_capacity.value: Unit.KN,
    ClusterResultColumn.net_design_bearing_capacity.value: Unit.KN,
    ClusterResultColumn.nominal_cpt.value: Unit.NONE,
    ClusterResultColumn.variation_coefficient.value: Unit.NONE,
    ClusterResultColumn.xi_factor.value: Unit.NONE,
    ClusterResultColumn.xi_values.value: Unit.NONE,
}


class CasesGrouperResults:
    """
    Container class for the results of multiple cases of GrouperResults.
    It expects that all GrouperResults objects are based on MultiCPTBearingResults with the same pile tip levels and test ids.
    """

    def __init__(
        self,
        results_per_case: Dict[Hashable, GrouperResults],
        cpt_locations: Dict[str, Location],
    ) -> None:
        """
        Initialize the class with the results of multiple cases of GrouperResults.

        Parameters
        ----------
        results_per_case : Dict[Hashable, GrouperResults]
            A dictionary with the results of multiple cases of GrouperResults.
            The keys of the dictionary are the case names.
            All GrouperResults objects must be generated with MultiCPTBearingResults objects that have the same pile tip levels and test ids.
        cpt_locations : Dict[str, Location]
            The mapping between `test_id` (key) and the `location` (value) of the cpt.
            The keys must contain all the `test_id` (s) used in the GrouperResults objects.

        Raises
        ------
        TypeError
            If `results_per_case` is not of the expected type.
            If `cpt_locations` is not of the expected type.
        ValueError
            If `results_per_case` is an empty dictionary.
            If not all GrouperResults objects have the same pile tip levels and test ids.
            If not all the `test_id` (s) used in the GrouperResults objects are in the keys of `cpt_locations`.
        """
        _validate_results_per_case(results_per_case)

        self._results_per_case = results_per_case
        self._cases = natsorted(list(results_per_case.keys()))
        self._grouper_results = [results_per_case[c] for c in self.cases]
        first_results = results_per_case[self.cases[0]]
        self._test_ids = natsorted(
            first_results.multi_cpt_bearing_results.cpt_results.test_ids
        )
        self._pile_tip_levels_nap = sorted(
            list(
                first_results.multi_cpt_bearing_results.cpt_results.to_pandas().pile_tip_level_nap.unique()
            ),
            reverse=True,
        )
        self._set_cpt_locations(cpt_locations)
        self._max_bearing_results_table = CptResultsTable.initialize()
        self._set_cpt_max_results_table(results_per_case)

    def _set_cpt_max_results_table(
        self, results_per_case: dict[Hashable, GrouperResults]
    ) -> None:
        """Private method to set the maximum results dataframe for each case."""

        for case_name, case_results in results_per_case.items():
            for cpt_results in case_results.cpt_results.results:
                test_id = str(cpt_results.soil_properties.test_id)
                for result_definition in GrouperResultsDefinition:
                    try:
                        result_values = getattr(
                            cpt_results.table, result_definition.value.name
                        )
                    except AttributeError:
                        continue

                    for idx, ptl_nap in enumerate(cpt_results.table.pile_tip_level_nap):
                        self._max_bearing_results_table.add_entry(
                            case_name=case_name,
                            result_def=result_definition.value,
                            test_id=test_id,
                            x=self.cpt_locations[test_id].x,
                            y=self.cpt_locations[test_id].y,
                            pile_tip_level_nap=ptl_nap,
                            result_value=result_values[idx],
                        )

    def _set_cpt_locations(self, value: Dict[str, Location]) -> None:
        """Private setter for `cpt_locations`."""
        if not isinstance(value, dict):
            raise TypeError(
                f"Expected type 'Dict[str, Location]' for 'cpt_locations', but got {type(value)}"
            )

        if not all(isinstance(k, str) for k in value.keys()):
            raise TypeError(
                f"Expected type 'str' for keys of 'cpt_locations', but got {(type(k) for k in value.keys())}"
            )

        if not all(isinstance(v, Location) for v in value.values()):
            raise TypeError(
                f"Expected type 'Location' for values of 'cpt_locations', but got {(type(v) for v in value.values())}"
            )

        if not all(test_id in value.keys() for test_id in self.test_ids):
            missing_test_ids = [
                test_id for test_id in self.test_ids if test_id not in value.keys()
            ]
            raise ValueError(
                "Not all `test_id`s used in the GrouperResults objects are in the keys of `cpt_locations`. "
                + f"The following `test_id`s are missing: {missing_test_ids}"
            )

        self._cpt_locations = value

    @property
    def results_per_case(self) -> Dict[Hashable, GrouperResults]:
        """The dictionary with case names as keys and GrouperResults as values."""
        return self._results_per_case

    @property
    def cases(self) -> List[Hashable]:
        """The case names."""
        return self._cases

    @property
    def grouper_results(self) -> List[GrouperResults]:
        """The GrouperResults objects."""
        return self._grouper_results

    @property
    def test_ids(self) -> List[str]:
        """The test ids of all GrouperResults objects."""
        return self._test_ids

    @property
    def pile_tip_levels_nap(self) -> List[float]:
        """The pile tip levels NAP of all GrouperResults objects."""
        return self._pile_tip_levels_nap

    @property
    def cpt_locations(self) -> Dict[str, Location]:
        """The mapping between `test_id` (key) and the `location` (value) of the cpt."""
        return self._cpt_locations

    @property
    def cpt_results_table(self) -> CptResultsTable:
        """Table object with the maximum bearing results of all cases."""
        return self._max_bearing_results_table


def _validate_results_per_case(
    results_per_case: Dict[Hashable, GrouperResults],
) -> None:
    """
    Private method to validate the results_per_case dictionary.

    Parameters
    ----------
    results_per_case : Dict[Hashable, GrouperResults]
        A dictionary with the results of multiple cases of GrouperResults.
        The keys of the dictionary are the case names.
        All GrouperResults objects must be generated with MultiCPTBearingResults objects that have the same pile tip levels and test ids.

    Raises
    ------
    TypeError
        If `results_per_case` is not of the expected type.
    ValueError
        If `results_per_case` is an empty dictionary.
        If not all GrouperResults objects have the same pile tip levels and test ids.
    """
    if not isinstance(results_per_case, dict):
        raise TypeError(
            f"Expected type 'Dict[Hashable, GrouperResults]' for 'results_per_case', but got {type(results_per_case)}"
        )

    if len(results_per_case) == 0:
        raise ValueError("Empty dictionary 'results_per_case' is not allowed.")

    for val in results_per_case.values():
        if not isinstance(val, GrouperResults):
            raise TypeError(
                f"Expected type 'GrouperResults' for items in 'results_per_case', but got {type(val)}"
            )

    first_key = list(results_per_case.keys())[0]
    test_ids = results_per_case[
        first_key
    ].multi_cpt_bearing_results.cpt_results.test_ids
    pile_tip_levels_nap = list(
        results_per_case[first_key]
        .multi_cpt_bearing_results.cpt_results.to_pandas()
        .pile_tip_level_nap.unique()
    )
    for results in results_per_case.values():
        if results.multi_cpt_bearing_results.cpt_results.test_ids != test_ids:
            raise ValueError("All GrouperResults objects must have the same test ids.")

        if (
            list(
                results.multi_cpt_bearing_results.cpt_results.to_pandas().pile_tip_level_nap.unique()
            )
            != pile_tip_levels_nap
        ):
            raise ValueError(
                "All GrouperResults objects must have the same pile tip levels."
            )



from pypilecore.viewers.viewer_cpt_results_overview import ViewerCptResultsOverview
from pypilecore.viewers.viewer_cpt_results_plan_view import ViewerCptResultsPlanView
from pypilecore.viewers.viewer_cpt_results_versus_ptls import ViewerCptResults
from pypilecore.viewers.viewer_grouper_results_per_cpt_table import (
    ViewerGrouperResultsPerCptTable,
)
from pypilecore.viewers.viewer_results_per_cpt_table import (
    ViewerResultsPerCptTable,
)

__all__ = [
    "ViewerCptResults",
    "ViewerCptResultsPlanView",
    "ViewerResultsPerCptTable",
    "ViewerGrouperResultsPerCptTable",
    "ViewerCptResultsOverview",
]



from __future__ import annotations

import datetime
from copy import deepcopy
from typing import Any, Dict, List, Literal, Mapping, Sequence, Tuple

from pygef.cpt import CPTData

from pypilecore.common.piles import PileGridProperties, PileProperties
from pypilecore.input.soil_properties import create_soil_properties_payload


def create_multi_cpt_payload(
    pile_tip_levels_nap: Sequence[float],
    cptdata_objects: List[CPTData],
    classify_tables: Dict[str, dict],
    groundwater_level_nap: float,
    pile: PileProperties,
    pile_grid: PileGridProperties | None,
    stiff_construction: bool = False,
    ocr: float | None = None,
    soil_load_sls: float = 0.0,
    pile_head_level_nap: float | Literal["surface"] = "surface",
    excavation_depth_nap: float | None = None,
    top_of_tension_zone_nap: float | None = None,
    individual_top_of_tension_zone_nap: Mapping[Any, float] | None = None,
    excavation_param_t: float = 1.0,
    excavation_stress_reduction_method: Literal["constant", "begemann"] = "constant",
    excavation_width: float | None = None,
    excavation_edge_distance: float | None = None,
    individual_ocr: Mapping[str, float] | None = None,
    overrule_xi: float | None = None,
    gamma_f_nk: float = 1.0,
    gamma_r_s: float = 1.2,
    gamma_r_b: float = 1.2,
    void_ratio_max: float = 0.8,
    void_ratio_min: float = 0.4,
    pile_load_sls_max: float = 1,
    pile_load_sls_min: float = 0,
    gamma_s_t: float = 1.35,
    gamma_gamma: float = 1.1,
    construction_sequence: Literal["cpt-pile", "pile-cpt"] = "cpt-pile",
) -> Tuple[dict, Dict[str, dict]]:
    """
    Creates a dictionary with the payload content for the PileCore endpoints
    "/tension/nen/multi-cpt/results" or "/tension/cur/multi-cpt/results"

    This dictionary can be passed directly to `nuclei.client.call_endpoint()`.

    Note
    ----
    The result should be converted to a jsonifyable message before it can be passed to a
    `requests` call directly, for instance with
    `nuclei.client.utils.python_types_to_message()`.

    Parameters
    ----------
    pile_tip_levels_nap:
        Range of values for which the pile resistance is calculated.
    cptdata_objects:
        A list of pygef.CPTData objects
    classify_tables:
        A dictionary, mapping `CPTData.alias` values to dictionary with the resulting response
        of a call to CPTCore `classify/*` information, containing the following keys:

            - geotechnicalSoilName: Sequence[str]
                Geotechnical Soil Name related to the ISO
            - lowerBoundary: Sequence[float]
                Lower boundary of the layer [m]
            - upperBoundary: Sequence[float]
                Upper boundary of the layer [m]
            - color: Sequence[str]
                Hex color code
            - mainComponent: Sequence[Literal["rocks", "gravel", "sand", "silt", "clay", "peat"]]
                Main soil component
            - cohesion: Sequence[float]
                Cohesion of the layer [kPa]
            - gamma_sat: Sequence[float]
                Saturated unit weight [kN/m^3]
            - gamma_unsat: Sequence[float]
                Unsaturated unit weight [kN/m^3]
            - phi: Sequence[float]
                Phi [degrees]
            - undrainedShearStrength: Sequence[float]
                Undrained shear strength [kPa]

    groundwater_level_nap:
        The groundwater level. Unit: [m] w.r.t. NAP.
    pile:
        A PileProperties object.
    stiff_construction:
        Set to True if it's a stiff costruction. This will have influence on the xi factor
        if you don't overrule it. Default = False.
    ocr:
        The Over-Consolidation-Ratio [-] of the foundation layer.
    soil_load_sls:
        Load on soil surface [kPa], used to calculate soil settlement. This is only required
        with the settlement-driven friction-range strategy.
        Default = 0.0
    pile_head_level_nap:
        The level of the pile head. Can be:

            - float;
                This is interpreted as an absolute level in [m w.r.t. NAP];
            - Literal["surface"] (default).
                In this case, the soil_properties.service_level
                property is used.
    top_of_tension_zone_nap:
        The level [w.r.t. NAP] of the top of the tension zone from which the uplift capacity
        (tension) is calculated. If None, the top of the tension zone is set to the `pile_head_level_nap`.
        Note that PileCore will automatically apply the remarks regarding L;a in CUR236 6.1.1. if
        this standard is selected.
    individual_top_of_tension_zone_nap:
        A dictionary, mapping specific CPTs (using the `CPTData.alias`` as keys) to `top_of_tension_zone_nap`
        (as values). This will overrule the general `top_of_tension_zone_nap` setting for these specific
        CPTs only.
        Example: {"S03": 1.5}
    individual_ocr:
        A dictionary, mapping specific CPTs (using the `CPTData.alias`` as keys) to Over-Consolidation-Ratio [-]
        of the foundation layer (as values). This will overrule the general `ocr` setting for
        these specific CPTs only.
        Example: {"S03": 1.5}
    excavation_depth_nap:
        Soil excavation depth after the CPT was taken. Unit: [m] w.r.t. NAP.
    excavation_param_t:
        Excavation parameter depending on pile installation method and/or phasing.
            - Use 1.0 (default) for post-excavation installation with vibration (i.e. hammering).
            - Use 0.5 for reduced-vibration installation, or pile installation prior to
              excavation method.

        See for more info NEN 9997-1+C2:2017 7.6.2.3.(10)(k)
    excavation_stress_reduction_method:
        Method used to calculate the stress reduction due to the excavation applied to the effective and total stresses.
        Only used when `excavation_depth_nap` is different than `None`. It can be:
            - "constant" (default): The stress reduction below the excavation is constant with depth. The stress reduction
            is equal to the original effective stress (i.e. before the excavation) at the excavation depth.
            - "begemann": The stress reduction below the excavation decreases with depth according to the Begemann method.
            This method uses the elastic solution of a strip load acting on a semi-infinite homogeneous soil mass.
            The load has a width equal to the excavation width and a magnitude equal to the original effective stress at
            the excavation depth.
        Regardless the method, the stress reduction applied above the excavation is equal to the original effective stress
        at each corresponding depth.
    excavation_width:
        Width of the excavation [m]. Used to calculate the stress reduction due to the excavation if the Begemann method is selected.
        In this case, it must be provided and it must be > 0.
    excavation_edge_distance:
        Distance from the pile centerline to the excavation edge [m]. Used to calculate the stress reduction due to the excavation if
        the Begemann method is selected. In this case, it must be provided and it must be between 0.0 and 0.5 * excavation_width.

        Note that:
            - 0.0 means that the pile is located at the edge of the excavation.
            - 0.5 * excavation_width means that the pile is at the center of the excavation.
    overrule_xi:
        Set a fixed value for xi in all calculations. Use with caution. This will
        overrule the calculation of xi-values based on the group-size, variation-
        coefficient and construction stiffness.
    gamma_f_nk:
        Safety factor for design-values of the negative sleeve friction force.
        Default = 1.0
    gamma_r_s:
        Safety factor, used to obtain design-values of the pile-tip bearingcapacity.
        Default = 1.2
    gamma_r_b:
        Safety factor, used to obtain design-values of the sleeve bearingcapacity.
        Default = 1.2
    void_ratio_max:
        Maximum void ratio of the soil (the loosest packing). The influence of this
        parameter is limited and therefore it is typically sufficient to provide a
        global estimation. For normally consolidated sands in The Netherlands, an
        emax = 0.80 can be used in most cases.
        Default = 0.8
    void_ratio_min:
        Minimum void ratio of the soil (the densest packing). The influence of this parameter
        is limited and therefore it is typically sufficient to provide a global estimation.
        For normally consolidated sands in The Netherlands, an emin = 0.40 can be used in most cases.
        Default = 0.4
    pile_load_sls_max:
        Maximum tension force on pile Ft;max;k. Note that only positive values (tension force)
        are accepted. unit: kN
        Default = 1
    pile_load_sls_min:
        Minimum tension force (or maximum compression force) on pile -Ft;min;k (tension > 0). Positive
        values (tension force), negative values (compression force) and 0.0 are accepted. Note that
        the positive must be <= pile_load_sls_max. unit: kN
        Default = 0
    gamma_s_t:
        Pile resistance factor gamma_s;t used to compute the design cone resistance values
        qc;z;d as prescribed in NEN 9997-1+C2_2017 7.6.3.3(d).
        Default = 1.35
    gamma_gamma:
        Partial factor for volumetric weight NEN 9997-1+C2:2017 A.3.2
        Default = 1.1
    construction_sequence:
        Value that indicates if CPT are performed before or after pile
        installation according to 7.3.1 CUR 2001-4.
        Default = "cpt-pile"

        Notes:
            - If standard "NEN9997-1" is used the following applies:
                - `cpt-pile`, then the `f1` factor is computed according to NEN 9997-1+C2:2017 7.6.3.3(e).
                - `pile-cpt`, then the `f1` factor is 1.0 at all depths.
            - If standard "CUR236" is used `f1` is always 1.0 regardless of the construction sequence.

    Returns
    -------
    multi_cpt_payload:
        Dictionary with the payload content for the PileCore endpoint
        "/tension/nen/multi-cpt/results" or "/tension/cur/multi-cpt/results" # TODO: check if valid for both
    results_kwargs:
        Dictionary with keyword arguments for the `pilecore.MultiCPTTensionBearingResults`
        object.

    Raises
    ------
    ValueError:
        - If `excavation_depth_nap` is not None and `excavation_param_t` is None.
        - If both `relative_pile_load` and `pile_load_sls` are None.
    """
    # Input validation
    if excavation_depth_nap is not None and excavation_param_t is None:
        raise ValueError(
            "`excavation_param_t` cannot be None when `excavation_depth_nap` is not None."
        )

    soil_properties_list, results_kwargs = create_soil_properties_payload(
        cptdata_objects=cptdata_objects,
        classify_tables=classify_tables,
        groundwater_level_nap=groundwater_level_nap,
        friction_range_strategy="manual",
        excavation_depth_nap=excavation_depth_nap,
        master_ocr=ocr,
        individual_negative_friction_range_nap=None,
        individual_positive_friction_range_nap=None,
        individual_ocr=individual_ocr,
        master_top_of_tension_zone_nap=top_of_tension_zone_nap,
        individual_top_of_tension_zone_nap=individual_top_of_tension_zone_nap,
    )
    multi_cpt_payload: dict = dict(
        pile_tip_levels_nap=list(pile_tip_levels_nap),
        list_soil_properties=soil_properties_list,
        pile_properties=pile.serialize_payload(),
        pile_head_level_nap=pile_head_level_nap,
        stiff_construction=stiff_construction,
        soil_load=soil_load_sls if soil_load_sls is not None else 0.0,
        excavation_param_t=excavation_param_t,
        gamma_f_nk=gamma_f_nk,
        gamma_r_b=gamma_r_b,
        gamma_r_s=gamma_r_s,
        void_ratio_max=void_ratio_max,
        void_ratio_min=void_ratio_min,
        pile_load_sls_max=pile_load_sls_max,
        pile_load_sls_min=pile_load_sls_min,
        gamma_s_t=gamma_s_t,
        gamma_gamma=gamma_gamma,
        construction_sequence=construction_sequence,
    )

    # Add optional properties
    if excavation_depth_nap is not None:
        multi_cpt_payload["excavation_depth_nap"] = excavation_depth_nap
    if excavation_stress_reduction_method == "constant":
        multi_cpt_payload["excavation_settings"] = dict(
            stress_reduction_method="constant"
        )
    else:
        multi_cpt_payload["excavation_settings"] = dict(
            stress_reduction_method="begemann",
            excavation_width=excavation_width,
            excavation_edge_distance=excavation_edge_distance,
        )
    if overrule_xi is not None:
        multi_cpt_payload["overrule_xi"] = overrule_xi
    if pile_grid is not None:
        multi_cpt_payload["pile_grid"] = pile_grid.serialize_payload()

    return multi_cpt_payload, results_kwargs


def create_multi_cpt_report_payload(
    multi_cpt_payload: dict,
    project_name: str,
    project_id: str,
    author: str,
    date: str | None = None,
    group_results_content: bool = True,
    individual_cpt_results_content: bool = True,
    result_summary_content: bool = True,
) -> dict:
    """
    Creates a dictionary with the payload content for the PileCore endpoint
    "/compression/multi-cpt/report"

    This dictionary can be passed directly to `nuclei.client.call_endpoint()`. Note that
    it should be converted to a jsonifyable message before it can be passed to a
    `requests` call directly, for instance with
    `nuclei.client.utils.python_types_to_message()`.

    Parameters
    ----------
    multi_cpt_payload:
        Index 0 of the resulting tuple of a call to `create_multi_cpt_payload()`
    project_name:
        The name of the project.
    project_id:
        The identifier (code) of the project.
    author:
        The author of the report.
    date:
        The date of the report. If None, the current date will be used.
    group_results_content:
        Whether or not to add the group-results section to the report. Default = True
    individual_cpt_results_content:
        Whether or not to add the individual-cpt-results section to the report.
        Default = True
    result_summary_content:
        Whether or not to add the result-summary section to the report. Default = True

    Returns
    -------
    report_payload:
        Dictionary with the payload content for the PileCore endpoint
        "/compression/multi-cpt/report"
    """
    report_payload = deepcopy(multi_cpt_payload)
    report_payload.update(
        dict(
            content=dict(
                group_results=group_results_content,
                individual_cpt_results=individual_cpt_results_content,
                result_summary=result_summary_content,
            ),
            general=dict(
                author=author,
                project_id=project_id,
                project_name=project_name,
                date=(
                    date if date is None else datetime.date.today().strftime("%d-%m-%y")
                ),
            ),
        )
    )
    return report_payload



from __future__ import annotations  # noqa: F404

from typing import Hashable, List

import pandas as pd
import plotly.express as px
import plotly.graph_objects as go

from pypilecore.results.cases_grouper_results import CasesGrouperResults
from pypilecore.results.data_tables import ResultsPandasColumn
from pypilecore.results.result_definitions import CPTResultDefinition
from pypilecore.results.typing import CasesMultiCPTResultsLike


class FigureCPTResultsVersusPtls:
    """
    Interactive figure to show the CPT results of the bearing capacity calculations
    versus the pile tip levels (PTLs).

    The layout of the figure is:
        - X axis: result values.
        - Y axis: pile tip level w.r.t. NAP.
        - Each trace represents a different CPT.

    The figure has a method to switch between cases and results.
    """

    def __init__(self, results_cases: CasesMultiCPTResultsLike) -> None:
        """
        Initializes the figure.

        Parameters
        ----------
        results_cases : CasesMultiCPTResultsProtocol
            The case results of the bearing capacity calculations.
        """
        # Validate the data
        if not isinstance(results_cases, CasesMultiCPTResultsLike):
            raise TypeError(
                f"Incompatible type for 'cases_multi_results': {type(results_cases)}"
            )

        # Set attributes
        self._results = results_cases

        # Initialize the figure
        self._figure = go.FigureWidget()

    @property
    def results(self) -> CasesMultiCPTResultsLike:
        """The results of the bearing capacity calculations."""
        return self._results

    @property
    def data(self) -> pd.DataFrame:
        """The dataframe used to plot the results."""
        return self.results.cpt_results_table.to_pandas()

    @property
    def cases(self) -> List[Hashable]:
        """The case names of each MultiCPTBearingResults."""
        return self.results.cases

    @property
    def test_ids(self) -> List[str]:
        """The test_ids (cpt names) of all the MultiCPTBearingResults."""
        return self.results.test_ids

    @property
    def figure(self) -> go.FigureWidget:
        """The figure widget."""
        return self._figure

    def get_visible_test_ids(self) -> List[go.Scatter]:
        """Returns the visible `test_id` (s) in the figure widget."""
        return [trace.name for trace in self.figure.data if trace.visible is True]

    def show_case_and_result(self, case_name: Hashable, result_name: str) -> None:
        """Shows the results for all CPTs and pile tip levels for the requested `case_name` and `result_name`.

        Parameters
        ----------
        case_name : str
            The name of the case to show.
        result_name : str
            The name of the result to show.

        Raises
        ------
        ValueError
            If the `case_name` is not found in the cases.
            If the `result_name` is not found in the CPTResultDefinitions.
        """
        # Check that case name is in cases.
        if case_name not in self.cases:
            raise ValueError(f"Case name '{case_name}' not found in cases.")

        # Get the result definition that corresponds to the result name.
        result_definition = CPTResultDefinition.get(result_name)

        # Get the visible test_ids
        if len(self.figure.data) == 0:
            # Select all test_ids if there are no traces in the figure yet.
            visible_test_ids = self.test_ids
        else:
            visible_test_ids = self.get_visible_test_ids()

        # Select data for case name and result name.
        mask_case_name = (
            self.data[ResultsPandasColumn.CASE_NAME.value] == case_name
            if case_name is not None
            else self.data[ResultsPandasColumn.CASE_NAME.value].isna()
        )
        selected_data = self.data.loc[
            (mask_case_name)
            & (
                self.data[ResultsPandasColumn.RESULT_NAME.value]
                == result_definition.name
            )
        ]
        traces = []
        for test_id in self.test_ids:
            df = selected_data.loc[
                selected_data[ResultsPandasColumn.TEST_ID.value] == test_id
            ]
            traces.append(
                go.Scatter(
                    x=df[ResultsPandasColumn.RESULT.value],
                    y=df[ResultsPandasColumn.PILE_TIP_LEVEL_NAP.value],
                    mode="lines+markers",
                    name=test_id,
                )
            )

        with self.figure.batch_update():
            # Empty traces
            self.figure.data = []

            # Apply changes
            self.figure.add_traces(traces)

            calculation_type = (
                "Grouper" if isinstance(self.results, CasesGrouperResults) else ""
            )

            self.figure.update_layout(
                title="CPT "
                + calculation_type
                + f"Results vs. Pile tip level<br>Case: {case_name}, Result: {result_definition.value.html}",
                xaxis=go.layout.XAxis(
                    title=f"{result_definition.value.html} [{result_definition.value.unit}]",
                    title_font_size=18,
                ),
                showlegend=True,
                # The following parameters shouldn't be necessary to update
                # (and therefore could be assigned on initialization of self.figure),
                # but they are needed to avoid a bug in plotly since dash > 3.0.
                height=800,
                width=800,
                legend_title="CPT",
                colorway=px.colors.qualitative.Plotly,
                yaxis=go.layout.YAxis(
                    title="Pile tip level [m NAP]",
                    title_font_size=18,
                ),
                autosize=False,
            )

            self.figure.update_traces(
                selector=lambda x: x.name in visible_test_ids,
                patch=dict(
                    visible=True,
                ),
            )

            self.figure.update_traces(
                selector=lambda x: x.name not in visible_test_ids,
                patch=dict(
                    visible="legendonly",
                ),
            )



from typing import Dict, Hashable, List, Protocol, Union, runtime_checkable

from pypilecore.common.piles import PileProperties
from pypilecore.results.compression.multi_cpt_results import (
    CPTCompressionGroupResultsTable,
    SingleCPTCompressionBearingResultsContainer,
)
from pypilecore.results.data_tables import CptResultsTable
from pypilecore.results.post_processing import MaxBearingResults
from pypilecore.results.tension.multi_cpt_results import (
    CPTTensionGroupResultsTable,
    SingleCPTTensionBearingResultsContainer,
)


@runtime_checkable
class LikeMultiCPTResults(Protocol):
    """
    Protocol class for MultiCPTBearingResults from compression or tension endpoint response
    """

    def __init__(self) -> None: ...

    @property
    def pile_properties(self) -> PileProperties: ...

    @property
    def cpt_results(
        self,
    ) -> Union[
        SingleCPTCompressionBearingResultsContainer,
        SingleCPTTensionBearingResultsContainer,
        MaxBearingResults,
    ]: ...

    @property
    def cpt_names(self) -> List[str]: ...

    @property
    def group_results_table(
        self,
    ) -> Union[CPTCompressionGroupResultsTable, CPTTensionGroupResultsTable]: ...


@runtime_checkable
class CasesMultiCPTResultsLike(Protocol):
    """Protocol for CasesMultiCPTResults."""

    @property
    def cases(self) -> List[Hashable]:
        """The case names of each MultiCPTBearingResults."""
        ...

    @property
    def test_ids(self) -> List[str]:
        """The test_ids (cpt names) of all the MultiCPTBearingResults."""
        ...

    @property
    def pile_tip_levels_nap(self) -> List[float]:
        """The pile tip levels w.r.t. NAP of all the MultiCPTBearingResults."""
        ...

    @property
    def cpt_results_table(self) -> CptResultsTable:
        """The Table object with all CPT results."""
        ...

    @property
    def results_per_case(self) -> Dict[Hashable, LikeMultiCPTResults]:
        """The dictionary with case names as keys and MultiCPTBearingResults as values."""
        ...



from typing import Any

import pandas as pd
from IPython.display import DisplayHandle, display
from ipywidgets import widgets
from natsort import natsorted

from pypilecore.results.cases_multi_cpt_results import CasesMultiCPTBearingResults


class ViewerResultsPerCptTable:
    def __init__(self, cases_cpt_results: CasesMultiCPTBearingResults) -> None:
        self.cases_cpt_results = cases_cpt_results
        self._df_widget = widgets.Output()

        self._case_dropdown = widgets.Dropdown(
            description="Case:",
            value=cases_cpt_results.cases[0],
            options=cases_cpt_results.cases,
        )

        # Create a sorted list of unique result names
        _options = natsorted(
            set(
                [
                    result_def.name
                    for result_def in cases_cpt_results.cpt_results_table.result_def
                ]
            )
        )
        _initial_option = "R_c_d_net" if "R_c_d_net" in _options else _options[0]
        self._result_dropdown = widgets.Dropdown(
            description="Result:",
            value=_initial_option,
            options=_options,
        )

        # Initialize the dataframe
        self._update_case_result(None)

        # Set up callbacks
        self._case_dropdown.observe(self._update_case_result, "value")
        self._result_dropdown.observe(self._update_case_result, "value")

        # Set up layout
        self._control_widgets = widgets.HBox(
            [
                self._case_dropdown,
                self._result_dropdown,
            ]
        )
        self._layout = widgets.VBox([self._control_widgets, self._df_widget])

    def to_pandas(self) -> pd.DataFrame:
        """Return the currently displayed table as a pandas DataFrame."""
        return self.cases_cpt_results.results_per_case[
            self._case_dropdown.value
        ].cpt_results.get_results_per_cpt(self._result_dropdown.value)

    def _update_case_result(self, change: Any) -> None:
        """
        Private method to update the figure when the case, result name or pile tip level
        are changed in the control widgets.
        """
        with self._df_widget:
            self._df_widget.clear_output()
            display(self.to_pandas().round(1))

    def display(self) -> DisplayHandle | None:
        """Display the figure."""
        return display(self._layout)



from __future__ import annotations

from typing import Dict


class PileType:
    """The PileType class represents the type of a pile."""

    def __init__(
        self,
        reference: str | None = None,
        installation_method: str | None = None,
        is_prefab: bool | None = None,
        is_open_ended: bool | None = None,
        is_low_vibrating: bool | None = None,
        is_auger: bool | None = None,
        settlement_curve: int | None = None,
        alpha_s_sand: float | None = None,
        alpha_s_clay: dict | None = None,
        alpha_p: float | None = None,
        alpha_t_sand: float | None = None,
        alpha_t_clay: dict | None = None,
        negative_fr_delta_factor: float | None = None,
        adhesion: float | None = None,
        qc_z_a_lesser_1m: float | None = None,
        qc_z_a_greater_1m: float | None = None,
        qb_max_limit: float | None = None,
        chamfered: float | None = None,
    ):
        """
        Represents the type of a pile.

        Parameters:
        -----------
        reference : str, optional
            The standard pile definition of the pile type (if applicable), by default None.
        installation_method: str, optional
            The installation method of the pile, by default None.
        alpha_s_sand : float, optional
            The alpha_s_sand value of the pile type, by default None.
        alpha_s_clay : float, optional
            The alpha_s_clay value of the pile type, by default None.
        alpha_p : float, optional
            The alpha_p value of the pile type, by default None.
        alpha_t_sand : float, optional
            The alpha_t_sand value of the pile type, by default None.
        alpha_t_clay : float, optional
            The alpha_t_clay value of the pile type, by default None.
        settlement_curve : int, optional
            The settlement curve of the pile type, by default None.
        negative_fr_delta_factor : float, optional
            The negative_fr_delta_factor value of the pile type, by default None.
        adhesion : float, optional
            The adhesion value of the pile type, by default None.
        is_low_vibrating : bool, optional
            The is_low_vibrating value of the pile type, by default None.
        is_auger : bool, optional
            The is_auger value of the pile type, by default None.
        qc_z_a_lesser_1m : float, optional
            Maximum cone resistance qc value allowed for layers with thickness < 1m in the calculation of positive skin friction resistance.
            It must be less or equal to `qc_z_a_greater_1m`.
            If None, then 12 MPa is used.
            Default: None.
        qc_z_a_greater_1m : float, optional
            Maximum cone resistance qc value allowed for layers with thickness >= 1m in the calculation of positive skin friction resistance.
            It must be greater or equal to `qc_z_a_lesser_1m`.
            If None, then 15 MPa is used.
            Default: None.
        qb_max_limit : float, optional
            Maximum value allowed for the pile tip resistance qb_max.
            Note that is value will be used as a limit for qb_max unless use_almere_rules is true and a higher value than 12 MPa is specified.
            If None, then 15 MPa is used.
            Default: None.
        chamfered : float, optional
            The chamfered value of the pile type, by default None.
        """
        self._standard_pile = None if reference is None else {"reference": reference}
        self._reference = reference
        self._installation_method = installation_method
        self._is_prefab = is_prefab
        self._is_open_ended = is_open_ended
        self._alpha_s_sand = alpha_s_sand
        self._alpha_s_clay = alpha_s_clay
        self._alpha_p = alpha_p
        self._alpha_t_sand = alpha_t_sand
        self._alpha_t_clay = alpha_t_clay
        self._settlement_curve = settlement_curve
        self._negative_fr_delta_factor = negative_fr_delta_factor
        self._adhesion = adhesion
        self._is_low_vibrating = is_low_vibrating
        self._is_auger = is_auger
        self._qc_z_a_lesser_1m = qc_z_a_lesser_1m
        self._qc_z_a_greater_1m = qc_z_a_greater_1m
        self._qb_max_limit = qb_max_limit
        self._chamfered = chamfered

    @classmethod
    def from_api_response(cls, pile_type: dict) -> PileType:
        """
        Instantiates a PileType from a pile type object in the API response payload.

        Parameters:
        -----------
        pile_type: dict
            A dictionary that is retrieved from the API response payload at "/pile_properties/pile_type".

        Returns:
        --------
        PileType
            A pile type.
        """
        standard_pile = pile_type.get("standard_pile", {})
        properties = pile_type.get("properties", {})
        return cls(
            reference=standard_pile.get("reference"),
            is_prefab=properties.get("is_prefab"),
            is_open_ended=properties.get("is_open_ended"),
            is_low_vibrating=properties.get("is_low_vibrating"),
            is_auger=properties.get("is_auger"),
            installation_method=properties.get("installation_method"),
            alpha_s_sand=properties.get("alpha_s_sand"),
            alpha_s_clay=properties.get("alpha_s_clay"),
            alpha_p=properties.get("alpha_p"),
            alpha_t_sand=properties.get("alpha_t_sand"),
            alpha_t_clay=properties.get("alpha_t_clay"),
            settlement_curve=properties.get("settlement_curve"),
            negative_fr_delta_factor=properties.get("negative_fr_delta_factor"),
            adhesion=properties.get("adhesion"),
            qc_z_a_lesser_1m=properties.get("qc_z_a_lesser_1m"),
            qc_z_a_greater_1m=properties.get("qc_z_a_greater_1m"),
            qb_max_limit=properties.get("qb_max_limit"),
        )

    @property
    def reference(self) -> str | None:
        """The standard pile definition of the pile type (if applicable)"""
        return self._reference

    @property
    def alpha_s_sand(self) -> float | None:
        """The alpha_s_sand value of the pile type"""
        return self._alpha_s_sand

    @property
    def alpha_s_clay(self) -> dict | None:
        """The alpha_s_clay value of the pile type"""
        return self._alpha_s_clay

    @property
    def alpha_p(self) -> float | None:
        """The alpha_p value of the pile type"""
        return self._alpha_p

    @property
    def alpha_t_sand(self) -> float | None:
        """The alpha_t_sand value of the pile type"""
        return self._alpha_t_sand

    @property
    def alpha_t_clay(self) -> dict | None:
        """The alpha_t_clay value of the pile type"""
        return self._alpha_t_clay

    @property
    def settlement_curve(self) -> int | None:
        """The settlement curve of the pile type"""
        return self._settlement_curve

    @property
    def negative_fr_delta_factor(self) -> float | None:
        """The negative_fr_delta_factor value of the pile type"""
        return self._negative_fr_delta_factor

    @property
    def adhesion(self) -> float | None:
        """The adhesion value of the pile type"""
        return self._adhesion

    @property
    def is_prefab(self) -> bool | None:
        """The is_prefab value of the pile type"""
        return self._is_prefab

    @property
    def is_open_ended(self) -> bool | None:
        """The is_open_ended value of the pile type"""
        return self._is_open_ended

    @property
    def installation_method(self) -> str | None:
        """The installation_method value of the pile type"""
        return self._installation_method

    @property
    def is_low_vibrating(self) -> bool | None:
        """The is_low_vibrating value of the pile type"""
        return self._is_low_vibrating

    @property
    def is_auger(self) -> bool | None:
        """The is_auger value of the pile type"""
        return self._is_auger

    @property
    def qc_z_a_lesser_1m(self) -> float | None:
        """Maximum cone resistance qc value allowed for layers with thickness < 1m in the calculation of positive skin friction resistance."""
        return self._qc_z_a_lesser_1m

    @property
    def qc_z_a_greater_1m(self) -> float | None:
        """Maximum cone resistance qc value allowed for layers with thickness >= 1m in the calculation of positive skin friction resistance."""
        return self._qc_z_a_greater_1m

    @property
    def qb_max_limit(self) -> float | None:
        """Maximum value allowed for the pile tip resistance qb_max."""
        return self._qb_max_limit

    @property
    def chamfered(self) -> float | None:
        """The chamfered value of the pile type"""
        return self._chamfered

    def serialize_payload(self) -> Dict[str, str | dict]:
        """
        Serialize the pile type to a dictionary payload for the API.

        Returns:
            A dictionary payload containing the standard pile (if set), alpha_s_sand, alpha_s_clay, alpha_p,
            alpha_t_sand, settlement curve, negative_fr_delta_factor, adhesion, is_low_vibrating, and is_auger.
        """
        payload: Dict[str, str | dict] = {}

        if self._standard_pile is not None:
            payload["standard_pile"] = self._standard_pile

        custom_type_properties: Dict[str, float | bool | dict | str] = {}

        if self.alpha_s_sand is not None:
            custom_type_properties["alpha_s_sand"] = self.alpha_s_sand

        if self.alpha_s_clay is not None:
            custom_type_properties["alpha_s_clay"] = self.alpha_s_clay

        if self.alpha_p is not None:
            custom_type_properties["alpha_p"] = self.alpha_p

        if self.alpha_t_sand is not None:
            custom_type_properties["alpha_t_sand"] = self.alpha_t_sand

        if self.alpha_t_clay is not None:
            custom_type_properties["alpha_t_clay"] = self.alpha_t_clay

        if self.settlement_curve is not None:
            custom_type_properties["settlement_curve"] = self.settlement_curve

        if self.negative_fr_delta_factor is not None:
            custom_type_properties["negative_fr_delta_factor"] = (
                self.negative_fr_delta_factor
            )

        if self.adhesion is not None:
            custom_type_properties["adhesion"] = self.adhesion

        if self.installation_method is not None:
            custom_type_properties["installation_method"] = self.installation_method

        if self.is_prefab is not None:
            custom_type_properties["is_prefab"] = self.is_prefab

        if self.is_low_vibrating is not None:
            custom_type_properties["is_low_vibrating"] = self.is_low_vibrating

        if self.is_auger is not None:
            custom_type_properties["is_auger"] = self.is_auger

        if self.is_open_ended is not None:
            custom_type_properties["is_open_ended"] = self.is_open_ended

        if self.qc_z_a_lesser_1m is not None:
            custom_type_properties["qc_z_a_lesser_1m"] = self.qc_z_a_lesser_1m

        if self.qc_z_a_greater_1m is not None:
            custom_type_properties["qc_z_a_greater_1m"] = self.qc_z_a_greater_1m

        if self.qb_max_limit is not None:
            custom_type_properties["qb_max_limit"] = self.qb_max_limit

        if self.chamfered is not None:
            custom_type_properties["chamfered"] = self.chamfered

        if len(custom_type_properties.keys()) > 0:
            payload["custom_properties"] = custom_type_properties

        return payload



from typing import Any, Callable, Tuple

from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from pypilecore.results.soil_properties import get_soil_layer_handles


def plot_bearing_overview(
    plot_qc: Callable[[Axes | None, bool], Axes],
    plot_friction_ratio: Callable[[Axes | None, bool], Axes],
    plot_layers: Callable[[Axes | None, bool, bool], Axes],
    plot_bearing_capacities: Callable[[Axes | None, Tuple[float, float], bool], Axes],
    test_id: str | None = None,
    figsize: Tuple[float, float] = (10.0, 12.0),
    width_ratios: Tuple[float, float, float] = (1, 0.1, 2),
    add_legend: bool = True,
    **kwargs: Any,
) -> Figure:
    """
    Plot an overview of the bearing-capacities, including the .

    Parameters
    ----------
    plot_qc:
        Function to plot the qc values vs. depth.
    plot_friction_ratio:
        Function to plot the friction ratio vs. depth.
    plot_layers:
        Function to plot the soil layers vs. depth.
    plot_bearing_capacities:
        Function to plot the bearing capacities vs. depth.
    test_id:
        The test ID to display in the legend title.
    figsize:
        Size of the activate figure, as the `plt.figure()` argument.
    width_ratios:
        Tuple of width-ratios of the subplots, as the `plt.GridSpec` argument.
    add_legend:
        Add a legend to the second axes object
    **kwargs:
        All additional keyword arguments are passed to the `pyplot.subplots()` call.

    Returns
    -------
    fig:
        The matplotlib Figure
    """

    kwargs_subplot = {
        "gridspec_kw": {"width_ratios": width_ratios},
        "sharey": "row",
        "figsize": figsize,
        "tight_layout": True,
    }

    kwargs_subplot.update(kwargs)

    fig, _ = plt.subplots(
        1,
        3,
        **kwargs_subplot,
    )

    ax_qc, ax_layers, ax_bearing = fig.axes
    ax_rf = ax_qc.twiny()
    assert isinstance(ax_rf, Axes)

    # Plot bearing capacities
    plot_qc(ax_qc, False)
    plot_friction_ratio(ax_rf, False)
    plot_layers(ax_layers, False, False)

    # Set bounds of layers to qc plot
    bounds = ax_qc.get_ylim()
    ax_layers.set_ylim(bounds)

    plot_bearing_capacities(ax_bearing, figsize, False)

    if add_legend:
        ax_qc_legend_handles_list = ax_qc.get_legend_handles_labels()[0]
        ax_rf_legend_handles_list = ax_rf.get_legend_handles_labels()[0]
        ax_layers_legend_handles_list = get_soil_layer_handles()

        # Omit last 2 duplicate "bearing" handles
        # (groundwater_level and surface_level):
        ax_bearing_legend_handles_list = ax_bearing.get_legend_handles_labels()[0][2:]

        handles_list = [
            *ax_qc_legend_handles_list,
            *ax_rf_legend_handles_list,
            *ax_layers_legend_handles_list,
            *ax_bearing_legend_handles_list,
        ]

        ax_bearing.legend(
            handles=handles_list,
            loc="upper left",
            bbox_to_anchor=(1, 1),
            title=("name: " + test_id if test_id is not None else "name: unknown"),
        )

    return fig



from dataclasses import dataclass
from enum import Enum
from typing import Hashable

import pandas as pd

from pypilecore.results.result_definitions import ResultDefinition


class ResultsPandasColumn(Enum):
    CASE_NAME = "case_name"
    TEST_ID = "test_id"
    X = "x"
    Y = "y"
    PILE_TIP_LEVEL_NAP = "pile_tip_level_nap"
    RESULT_NAME = "result_name"
    RESULT = "result"
    RESULT_UNIT = "result_unit"


@dataclass
class CptResultsTable:
    case_name: list[Hashable]
    result_def: list[ResultDefinition]
    result_value: list[float]
    test_id: list[str]
    x: list[float]
    y: list[float]
    pile_tip_level_nap: list[float]

    @classmethod
    def initialize(cls) -> "CptResultsTable":
        return cls(
            case_name=[],
            result_def=[],
            result_value=[],
            test_id=[],
            x=[],
            y=[],
            pile_tip_level_nap=[],
        )

    def add_entry(
        self,
        case_name: Hashable,
        result_def: ResultDefinition,
        test_id: str,
        x: float | int,
        y: float | int,
        pile_tip_level_nap: float,
        result_value: float,
    ) -> None:
        """
        Add an data entry to the CPT results table.
        """
        self.case_name.append(case_name)
        self.result_def.append(result_def)
        self.test_id.append(test_id)
        self.x.append(x)
        self.y.append(y)
        self.pile_tip_level_nap.append(pile_tip_level_nap)
        self.result_value.append(result_value)

    def to_pandas(self) -> pd.DataFrame:
        """
        Return the CPT results as a pandas DataFrame.
        """
        return pd.DataFrame(
            {
                "case_name": self.case_name,
                "test_id": self.test_id,
                "x": self.x,
                "y": self.y,
                "pile_tip_level_nap": self.pile_tip_level_nap,
                "result_name": [rd.name for rd in self.result_def],
                "result": self.result_value,
                "result_unit": [rd.unit for rd in self.result_def],
            }
        )
